import os
import pandas as pd 

def saveDataframeCSV(dataframe: pd.DataFrame,
                     outputFilePath: str):
    """ Function to save a pandas Dataframe as a csv file with the index removed.
            Checks if the path in the outputFilePath exists and will create any missing directories.

    Parameters
    ----------
    dataframe : pd.DataFrame
        Pandas dataframe to save out as a csv
    outputFilePath : str
        Full file path to save the dataframe out to.
    """
    if not outputFilePath.endswith('.csv'):
        raise ValueError("This function saves .csv files, so outputFilePath must end in .csv")
    
    if not isinstance(dataframe, pd.DataFrame):
        raise ValueError("Function expects a pandas DataFrame to save out.")

    # Make directory if it doesn't exist
    if not os.path.exists(os.path.dirname(outputFilePath)):
        os.makedirs(os.path.dirname(outputFilePath))

    # Save out feature set
    dataframe.to_csv(outputFilePath, index=False)

    return


def matchCTtoSegmentation(imgFileListPath: str,
                          segType: str,
                          outputDirPath: str = None):
    """From full list of image files, extract CT and corresponding segmentation files and create new table.
    One row of the table contains both the CT and segmentation data for one patient.
    This function currently assumes there is one segmentation for each patient.

    Parameters
    ----------
    imgFileListPath : str
        Path to csv containing list of image directories/paths in the dataset. 
        Expecting output from med-imagetools autopipeline .imgtools_[dataset]
    segType : str
        Type of file segmentation is in. Can be SEG or RTSTRUCT.
    outputDirPath : str
        Optional path to directory to save the dataframe to as a csv.
    
    Returns
    -------
    pd.Dataframe
        Dataframe containing the CT and corresponding segmentation data for each patient

    Note: All subseries of CT will be kept in the dataframe in this function
    """
    # Check that segmentation file type is acceptable
    if segType != "RTSTRUCT" and segType != "SEG":
        raise ValueError("Incorrect segmentation file type. Must be RTSTRUCT or SEG.")

    # Check that imgFileListPath is a csv file to properly be loaded in
    if not imgFileListPath.endswith('.csv'):
        raise ValueError("This function expects to load in a .csv file, so imgFileListPath must end in .csv")

    # Load in complete list of patient image directories of all modalities (output from med-imagetools crawl)
    fullDicomList = pd.read_csv(imgFileListPath, index_col=0)

    # Extract all CT rows
    allCTRows = fullDicomList.loc[fullDicomList['modality'] == "CT"]

    # Extract all SEG rows
    allSegRows = fullDicomList.loc[fullDicomList['modality'] == segType]

    # Merge the CT and segmentation dataframes based on the CT ID (referenced in the segmentation rows)
    # Uses only segmentation keys, so no extra CTs are kept
    # If multiple CTs have the same ID, they are both included in this table
    samplesWSeg = allCTRows.merge(allSegRows, how='inner',
                                  left_on=['series', 'patient_ID'], 
                                  right_on=['reference_ct', 'patient_ID'], 
                                  suffixes=('_CT','_seg'))

    # Sort dataframe by ascending patient ID value
    samplesWSeg.sort_values(by='patient_ID', inplace=True)

    # Save out the combined list
    if outputDirPath != None:
        # Get datasetname from imgtools output
        datasetName = imgFileListPath.partition("imgtools_")[2]
        fileName = "ct_to_seg_match_list_" + datasetName
        # Join this name with the output directory and add file prefix and csv suffix
        outputFilePath = os.path.join(outputDirPath, fileName)
        saveDataframeCSV(samplesWSeg, outputFilePath)
    
    return samplesWSeg


def getSegmentationType(imgFileListPath: str):
    """Find the segmentation type from the full list of image files.

    Parameters
    ----------
    imgFileListPath : str
        Path to csv containing list of image directories/paths in the dataset. 
        Expecting output from med-imagetools autopipeline .imgtools_[dataset]

    Returns
    -------
    str
        Segmentation type (RTSTRUCT or SEG)
    """
    # Check that imgFileListPath is a csv file to properly be loaded in
    if not imgFileListPath.endswith('.csv'):
        raise ValueError("This function expects to load in a .csv file, so imgFileListPath must end in .csv")

    # Load in complete list of patient image directories of all modalities (output from med-imagetools crawl)
    fullDicomList = pd.read_csv(imgFileListPath, index_col=0)

    # Get list of unique modalities 
    modalities = list(fullDicomList['modality'].unique())

    if "RTSTRUCT" in modalities:
        segType = "RTSTRUCT"
    elif "SEG" in modalities:
        segType = "SEG"
    else:
        raise RuntimeError("No suitable segmentation type found. YAREA can only use RTSTRUCTs and DICOM-SEG segmentations.")

    return segType


import os
import pydicom
import SimpleITK as sitk

from imgtools.ops import StructureSetToSegmentation
from imgtools.io import read_dicom_auto


def loadDicomSITK(imgDirPath: str) -> sitk.Image:
    """Read DICOM series as SimpleITK Image.

    Parameters
    ----------
    img_path : str
        Path to directory containing the DICOM series to load.

    Returns
    -------
    sitk.Image
        The loaded image. 
    """
    # Set up the reader for the DICOM series
    reader = sitk.ImageSeriesReader()
    dicomNames = reader.GetGDCMSeriesFileNames(imgDirPath)
    reader.SetFileNames(dicomNames)
    return reader.Execute()


def loadRTSTRUCTSITK(rtstructPath: str,
                     baseImageDirPath: str,
                     roiNames:str = None) -> dict:
    """ Load RTSTRUCT into SimpleITK Image.

    Parameters
    ----------
    rtstructPath : str
        Path to the DICOM file containing the RTSTRUCT
    baseImageDirPath : str
        Path to the directory containing the DICOMS for the original image the segmentation 
        was created from (e.g. CT). This is required to load the RTSTRUCT.
    roiNames : str
        Identifier for which region(s) of interest to load from the total segmentation file
    
    Returns
    -------
    dict
        A dictionary of mask ROIs from the loaded RTSTRUCT image as a SimpleITK image objects.
        The segmentation label is set to 1.
    """

    # Set up segmentation loader
    makeMask = StructureSetToSegmentation(roi_names=roiNames)
    
    # Read in the base image (CT) and segmentation DICOMs into SITK Images
    baseImage = read_dicom_auto(baseImageDirPath)
    segImage = read_dicom_auto(rtstructPath)

    try:
        # Get the individual ROI masks
        segMasks = makeMask(segImage, baseImage.image, existing_roi_indices={"background":0}, ignore_missing_regex=False)
    except ValueError:
        return {}
        
    # Get list of ROIs present in this rtstruct
    loadedROINames = segMasks.raw_roi_names
    # Initialize dictionary to store ROI names and images
    roiStructs = {}
    # Get each roi and its label and store in dictionary
    for roi in loadedROINames:
        # Get the mask for this ROI
        roiMask = segMasks.get_label(name=roi)
        # Store the ROI name and image
        roiStructs[roi] = roiMask
    
    return roiStructs


def loadSegmentation(segImagePath: str,
                     modality: str,
                     baseImageDirPath: str = None,
                     roiNames: str = None) -> dict:
    ''' Function to load a segmentation with the correct function.
    
    Parameters
    ----------
    segImagePath : str
        Path to the segmentation file to load
    modality : str
        Type of image that imgPath points to to load. If RTSTRUCT, must set baseImageDirPath
    baseImageDirPath : str
        Path to the directory containing the DICOMS for the original image the segmentation 
        was created from. 
    roiNames : str
        Identifier for which region(s) of interest to load from the total segmentation file
        
    Returns
    -------
    dict
        A dictionary of each of the ROIs and their name in the segmentation image as sitk.Image objects.
    
    Examples
    --------
    >>> segImages = loadSegmentation("/path/to/segmentation/1.dcm", 'RTSTRUCT', '/path/to/CT', 'GTVp.*')
    '''

    if modality in ['SEG', 'seg']:
        # Loading SEG requires directory containing file, not the actual file path
        imgFolder, _ = os.path.split(segImagePath)
        segHeader = pydicom.dcmread(segImagePath, stop_before_pixels=True)
        roiName = segHeader.SegmentSequence[0].SegmentLabel
        return {roiName: loadDicomSITK(imgFolder)}
    
    elif modality in ['RTSTRUCT', 'rtstruct']:
        if baseImageDirPath == None:
            raise ValueError("Missing path to original image segmentation was taken from. RTSTRUCT loader requires original image.")
        else:
            return loadRTSTRUCTSITK(segImagePath, baseImageDirPath, roiNames)
    
    else:
        raise ValueError('This segmentation modality is not supported. Must be one of RTSTRUCT or SEG')




from imgtools.io import read_dicom_series
from itertools import chain
from joblib import Parallel, delayed
from radiomics import featureextractor, imageoperations, logging
import radiomics

from readii.image_processing import *
from readii.loaders import *
from readii.metadata import *
from readii.negative_controls import *

def singleRadiomicFeatureExtraction(ctImage:sitk.Image,
                                    roiImage:sitk.Image,
                                    pyradiomicsParamFilePath:str = "./src/readii/data/default_pyradiomics.yaml",
                                    negativeControl:str = None):
    """Function to perform radiomic feature extraction for a single CT image and its corresponding segmentation.
       CT and segmentation will be aligned and cropped prior to extraction. 

    Parameters
    ----------
    ctImage : sitk.Image
        CT image to perform feature extraction on. Will be cropped and potentially generate a negative control (see negativeControl arg)
    roiImage : sitk.Image
        Region of interest (ROI) to extract radiomic features from within the CT.
    pyradiomicsParamFilePath : str
        Path to file containing configuration settings for pyradiomics feature extraction. Will use the provided config file in 'data/' by default if no file passed in.
    negativeControl : str
        Name of negative control to generate from the CT to perform feature extraction on. If set to None, will extract features from original CT image.

    Returns
    -------
    dict
        Dictionary containing image metadata, versions for key packages used for extraction, and radiomic features
    """
    # If no pyradiomics paramater file passed, use default
    if pyradiomicsParamFilePath == None:
        pyradiomicsParamFilePath = "./src/readii/data/default_pyradiomics.yaml"

    # In case segmentation contains extra axis, flatten to 3D by removing it
    roiImage = flattenImage(roiImage)
    # Segmentation has different origin, align it to the CT for proper feature extraction
    alignedROIImage = alignImages(ctImage, roiImage)

    # Get pixel value for the segmentation
    segmentationLabel = getROIVoxelLabel(alignedROIImage)
    
    # Check that CT and segmentation correspond, segmentationLabel is present, and dimensions match
    segBoundingBox, correctedROIImage = imageoperations.checkMask(ctImage, alignedROIImage, label=segmentationLabel)
    # Update the ROI image if a correction was generated by checkMask
    if correctedROIImage is not None:
        alignedROIImage = correctedROIImage
    
    # Crop the image and mask to a bounding box around the mask to reduce volume size to process
    croppedCT, croppedROI = imageoperations.cropToTumorMask(ctImage, alignedROIImage, segBoundingBox)

    if negativeControl != None:
        print("Generating ", negativeControl, "negative control for CT.")
        # Make negative control version of ctImage
        croppedCT = applyNegativeControl(nc_type=negativeControl, baseImage=croppedCT, baseROI=croppedROI, roiLabel=segmentationLabel)

    # Load PyRadiomics feature extraction parameters to use
    # Initialize feature extractor with parameters
    try:
        featureExtractor = featureextractor.RadiomicsFeatureExtractor(pyradiomicsParamFilePath)
    except OSError as e:
        print("ERROR: Supplied pyradiomics parameter file does not exist or is not at that location.")
        raise

    # Extract radiomic features from CT with segmentation as mask
    idFeatureVector = featureExtractor.execute(croppedCT, croppedROI, label=segmentationLabel)

    return idFeatureVector


def radiomicFeatureExtraction(imageMetadataPath:str,
                              imageDirPath:str,
                              roiNames:str = None,
                              pyradiomicsParamFilePath:str = "src/readii/data/default_pyradiomics.yaml",
                              outputDirPath:str = None,
                              negativeControl:str = None,
                              parallel:bool = False,):
    """Perform radiomic feature extraction using PyRadiomics on CT images with a corresponding segmentation.
       Utilizes outputs from med-imagetools (https://github.com/bhklab/med-imagetools) run on the image dataset.

    Parameters
    ----------
    imageMetadataPath : str
        Path to csv file created by matchCTtoSegmentation function that contains a CT and matching segmentation in each row.
    imageDirPath : str
        Path to the directory containing the directory of CT and segmentation images. This directory should contain the .imgtools directory from the med-imagetools run
        and be the same as the input path used in med-imagetools
    roiNames : str
        Name pattern for the ROIs to load for the RTSTRUCTs. Can be None for DICOM SEG segmentations.
    pyradiomicsParamFilePath : str
        Path to file containing configuration settings for pyradiomics feature extraction. Will use the provided config file in 'data/' by default if no file passed in.
    outputDirPath : str
        Path to directory save the dataframe of extracted features to as a csv
    negativeControl : str
        Name of negative control to generate from the CT to perform feature extraction on. If set to None, will extract features from original CT image.
    parallel : bool
        Flag to decide whether to run extraction in parallel. 
    
    Returns
    -------
    pd.DataFrame
        Dataframe containing the image metadata and extracted radiomic features.
    """
    # Setting pyradiomics verbosity lower
    logger = logging.getLogger("radiomics")
    logger.setLevel(logging.ERROR)

    # If no pyradiomics paramater file passed, use default
    if pyradiomicsParamFilePath == None:
        pyradiomicsParamFilePath = "./src/readii/data/default_pyradiomics.yaml"

    # Load in summary file generated by radiogenomic_pipeline
    pdImageInfo = pd.read_csv(imageMetadataPath, header=0)

    # Get array of unique CT series' IDs to iterate over
    ctSeriesIDList = pdImageInfo["series_CT"].unique()

    def featureExtraction(ctSeriesID):
        ''' Function to extract PyRadiomics features for all ROIs present in a CT. Inner function so it can be run in parallel with joblib.'''
        # Get all info rows for this ctSeries
        ctSeriesInfo = pdImageInfo.loc[pdImageInfo["series_CT"] == ctSeriesID]
        patID = ctSeriesInfo.iloc[0]["patient_ID"]

        print("Processing ", patID)

        # Get absolute path to CT image files 
        ctDirPath = os.path.join(imageDirPath, ctSeriesInfo.iloc[0]['folder_CT'])
        # Load CT by passing in specific series to find in a directory
        ctImage = read_dicom_series(path = ctDirPath, series_id = ctSeriesID)

        # Get list of segmentations to iterate over
        segSeriesIDList = ctSeriesInfo['series_seg'].unique()

        # Initialize dictionary to store radiomics data for each segmentation (image metadata + features)
        ctAllData = []

        # Loop over every segmentation associated with this CT - only loading CT once
        for segCount, segSeriesID in enumerate(segSeriesIDList):
            segSeriesInfo = ctSeriesInfo.loc[ctSeriesInfo['series_seg'] == segSeriesID]

            # Check that a single segmentation file is being processed
            if len(segSeriesInfo) > 1:
                # Check that if there are multiple rows that it's not due to a CT with subseries (this is fine, the whole series is loaded)
                if not segSeriesInfo.duplicated(subset=['series_CT'], keep=False).all():
                    raise RuntimeError("Some kind of duplication of segmentation and CT matches not being caught. Check seg_and_ct_dicom_list in radiogenomic_output.")
            
            # Get absolute path to segmentation image file
            segFilePath = os.path.join(imageDirPath, segSeriesInfo.iloc[0]['file_path_seg'])
            # Get dictionary of ROI sitk Images for this segmentation file
            segImages = loadSegmentation(segFilePath, modality = segSeriesInfo.iloc[0]['modality_seg'], 
                                         baseImageDirPath = ctDirPath, roiNames = roiNames)
            
            # Check that this series has ROIs to extract from (dictionary isn't empty)
            if not segImages:
                print('CT ', ctSeriesID, 'and segmentation ',segSeriesID, ' has no ROIs or no ROIs with the label ', roiNames, '. Moving to next segmentation.')

            else:
                # Loop over each ROI contained in the segmentation to perform radiomic feature extraction
                for roiCount, roiImageName in enumerate(segImages):
                    # ROI counter for image metadata output
                    roiNum = roiCount + 1

                    # Extract features listed in the parameter file
                    print("Calculating radiomic features for segmentation:", roiImageName)

                    # Get sitk Image object for this ROI
                    roiImage = segImages[roiImageName]

                    # Exception catch for if the segmentation dimensions do not match that original image
                    try:
                        # Check if segmentation just has an extra axis with a size of 1 and remove it
                        if roiImage.GetDimension() > 3 and roiImage.GetSize()[3] == 1:
                            roiImage = flattenImage(roiImage)
                            
                        # Check that image and segmentation mask have the same dimensions
                        if ctImage.GetSize() != roiImage.GetSize():
                            # Checking if number of segmentation slices is less than CT 
                            if ctImage.GetSize()[2] > roiImage.GetSize()[2]:  
                                print("Slice number mismatch between CT and segmentation for", patID, ". Padding segmentation to match.")
                                roiImage = padSegToMatchCT(ctDirPath, segFilePath, ctImage, roiImage)
                            else:
                                raise RuntimeError("CT and ROI dimensions do not match.")

                    # Catching CT and segmentation size mismatch error
                    except RuntimeError as e:
                        print(str(e))


                    # Extract radiomic features from this CT/segmentation pair
                    idFeatureVector = singleRadiomicFeatureExtraction(ctImage, roiImage = roiImage,
                                                                      pyradiomicsParamFilePath = pyradiomicsParamFilePath,
                                                                      negativeControl = negativeControl)

                    # Create dictionary of image metadata to append to front of output table
                    sampleROIData = {"patient_ID": patID,
                                    "study_description": segSeriesInfo.iloc[0]['study_description_CT'],
                                    "series_UID": segSeriesInfo.iloc[0]['series_CT'],
                                    "series_description": segSeriesInfo.iloc[0]['series_description_CT'],
                                    "image_modality": segSeriesInfo.iloc[0]['modality_CT'],
                                    "instances": segSeriesInfo.iloc[0]['instances_CT'],
                                    "seg_series_UID": segSeriesInfo.iloc[0]['series_seg'],
                                    "seg_modality": segSeriesInfo.iloc[0]['modality_seg'],
                                    "seg_ref_image": segSeriesInfo.iloc[0]['reference_ct_seg'],
                                    "roi": roiImageName,
                                    "roi_number": roiNum,
                                    "negative_control": negativeControl}

                    # Concatenate image metadata with PyRadiomics features
                    sampleROIData.update(idFeatureVector)
                    # Store this ROI's info in the segmentation level list
                    ctAllData.append(sampleROIData)
                    
        return ctAllData
        ###### END featureExtraction #######

    # Extract radiomic features for each CT, get a list of dictionaries
    # Each dictioary contains features for each ROI in a single CT
    if not parallel:
        # Run feature extraction over samples in sequence - will be slower
        features = [featureExtraction(ctSeriesID) for ctSeriesID in ctSeriesIDList]
    else:
        # Run feature extraction in parallel
        features = Parallel(n_jobs=-1, require='sharedmem')(delayed(featureExtraction)(ctSeriesID) for ctSeriesID in ctSeriesIDList)
    
    # Flatten the list of dictionaries (happens when there are multiple ROIs or SEGs associated with a single CT)
    flatFeatures = list(chain.from_iterable(features))
    # Convert list of feature sets into a pandas dataframe to save out
    featuresTable = pd.DataFrame(flatFeatures)

    if outputDirPath != None:
        datasetName = imageMetadataPath.partition("match_list_")[2]
        # Setup output file name with the dataset name as a suffix
        if negativeControl == None:
            outFileName = "radiomicfeatures_" + datasetName
        else:
            # Add negative control identifier to output file name
            outFileName = "radiomicfeatures_" + negativeControl + "_" + datasetName

        # Join outputDirPath, a features directory, and the output file name
        outputFilePath = os.path.join(outputDirPath, "features/", outFileName)
        # Save out the features
        saveDataframeCSV(featuresTable, outputFilePath)

    return featuresTable


import SimpleITK as sitk
import numpy as np
import random

from readii.image_processing import alignImages


def shuffleImage(imageToShuffle: sitk.Image):
    """Function to shuffle all pixel values in a sitk Image (developed for 3D, should work on 2D as well)

    Parameters
    ----------
    imageToShuffle : sitk.Image
        Image to shuffle the pixels in

    Returns
    -------
    sitk.Image
        Image with all pixel values randomly shuffled with same dimensions as input image
    """
    # Convert the image to an array
    arrImage = sitk.GetArrayFromImage(imageToShuffle)

    # Get array dimensions to reshape back to
    imgDimensions = arrImage.shape

    # Flatten the 3D array to 1D so values can be shuffled
    flatArrImage = arrImage.flatten()

    # Shuffle the flat array
    np.random.shuffle(flatArrImage, )

    # Reshape the array back into the original image dimensions
    shuffled3DArrImage = np.reshape(flatArrImage, imgDimensions)

    # Convert back to sitk Image
    shuffledImage = sitk.GetImageFromArray(shuffled3DArrImage)

    # Set the origin/direction/spacing from original image to shuffled image
    alignedShuffledImage = alignImages(imageToShuffle, shuffledImage)

    return alignedShuffledImage


def makeRandomImage(baseImage: sitk.Image):
    """Function to generate random pixel values based on the range of values in a sitk Image (developed for 3D, should work on 2D as well)

    Parameters
    ----------
    baseImage : sitk.Image
        Image to randomly generate pixel values

    Returns
    -------
    sitk.Image
        Image with all pixel values randomly generated with same dimensions as input image
    """
    # Convert the image to an array
    arrImage = sitk.GetArrayFromImage(baseImage)

    # Get array dimensions to reshape back to
    imgDimensions = arrImage.shape

    # Get min and max HU values to set as range for random values
    minVoxelVal = np.min(arrImage)
    maxVoxelVal = np.max(arrImage)

    # Delete arrImage to save memory
    del (arrImage)

    # Generate random array with same dimensions as baseImage
    random3DArrImage = np.random.randint(low=minVoxelVal, high=maxVoxelVal, size=imgDimensions)

    # Convert random array to a sitk Image
    randomImage = sitk.GetImageFromArray(random3DArrImage)

    # Set the origin/direction/spacing from the original image to the random image
    alignedRandomImage = alignImages(baseImage, randomImage)

    return alignedRandomImage


def makeRandomRoi(baseImage: sitk.Image, baseROI: sitk.Image, roiLabel: int = 1):
    """Function to generate random pixel values within the Region of Interest based on the range of values in a sitk Image

    Parameters
    ----------
    baseImage : sitk.Image
        Image to randomly generate pixel values in Region of Interest
    baseROI : sitk.Image 
        Image detailing Region of Interest
    roiLabel : int
        The label representing the ROI in baseROI

    Returns
    -------
    sitk.Image
        Image with all pixel values within the Region of Interest randomly generated with same dimensions as input image
    """
    # Initialize variables to track the highest and lowest pixel values in the ROI
    maxVoxelVal = float('-inf')
    minVoxelVal = float('inf')

    # Iterate through baseROI to find the highest and lowest values in baseImage's ROI
    baseROISize = baseROI.GetSize()
    for x in range(baseROISize[0]):
        for y in range(baseROISize[1]):
            for z in range(baseROISize[2]):
                # Ensure only pixels that match the roiLabel of hte image are looked at
                if baseROI.GetPixel(x, y, z) == roiLabel:
                    # Once a pixel is confirmed to be in the ROI get it's corresponding value from the baseImage
                    current_value = baseImage.GetPixel(x, y, z)
                    # Update the max and min values
                    maxVoxelVal = max(maxVoxelVal, current_value)
                    minVoxelVal = min(minVoxelVal, current_value)

    # Create a new base image so we are not directly editing the input image
    new_base = baseImage.__copy__()
    # Delete the input image to save space
    del (baseImage)

    # Now iterate over the pixels of the ROI in the image and randomly generate a new value for them
    for x in range(baseROISize[0]):
        for y in range(baseROISize[1]):
            for z in range(baseROISize[2]):
                if baseROI.GetPixel(x, y, z) == roiLabel:
                    # Randomly assigning the current value to the range [maxVoxelVal, maxVoxelVal]
                    mapped_value = random.randint(minVoxelVal, maxVoxelVal)

                    # Set the new pixel value
                    new_base.SetPixel(x, y, z, mapped_value)

    return new_base


def shuffleROI(baseImage: sitk.Image, baseROI: sitk.Image, roiLabel: int = 1):
    """Function to shuffle all pixel values within the Region of Interest in a sitk Image

    Parameters
    ----------
    baseImage : sitk.Image
        Image to shuffle the Region of Interest pixels in
    baseROI : sitk.Image
        Image detailing Region of Interest
    roiLabel : int
        The label representing the ROI in baseROI

    Returns
    -------
    sitk.Image
        Image with all pixel values in the Region of Interest randomly shuffled with same dimensions as input image
    """

    # A collection of corresponding value in the BaseImage of all the pixels in the ROI
    count = []
    # Iterate through baseROI to store the corresponding value in the baseImage of all the pixels in the ROI
    baseROISize = baseROI.GetSize()
    for x in range(baseROISize[0]):
        for y in range(baseROISize[1]):
            for z in range(baseROISize[2]):
                if baseROI.GetPixel(x, y, z) == roiLabel:
                    # append the ROI corresponding pixel values to the list
                    count.append(baseImage.GetPixel(x, y, z))

    # Create a new base image so we are not directly editing the input image
    new_base = baseImage.__copy__()
    # Delete the input image to save space
    del (baseImage)

    # # Randomly shuffling the pixel values
    random.shuffle(count)

    for x in range(baseROISize[0]):
        for y in range(baseROISize[1]):
            for z in range(baseROISize[2]):
                if baseROI.GetPixel(x, y, z) == roiLabel:
                    # Set the value of a pixel in the ROI to be a shuffled value
                    new_base.SetPixel(x, y, z, count.pop())

    return new_base


def makeRandomNonRoi(baseImage: sitk.Image, baseROI: sitk.Image, roiLabel: int = 1):
    """Function to generate random pixel values outside the Region of Interest based on the range of values in a sitk Image

    Parameters
    ----------
    baseImage : sitk.Image
        Image to randomly generate pixel values in outside the Region of Interest
    baseROI : sitk.Image
        Image detailing the Region of Interest
    roiLabel : int 
        The label representing the ROI in baseROI

    Returns
    -------
    sitk.Image
        Image with all pixel values outside the Region of Interest randomly generated with same dimensions as input image
    """
    # Initialize variables to track the highest and lowest pixel values not in the ROI
    maxVoxelVal = float('-inf')
    minVoxelVal = float('inf')

    # Iterate through baseImage to find the highest and lowest values not in baseImage's ROI
    baseImageSize = baseImage.GetSize()
    baseROISize = baseROI.GetSize()
    for x in range(baseImageSize[0]):
        for y in range(baseImageSize[1]):
            for z in range(baseImageSize[2]):
                # Ensure only pixels that do not match the roiLabel of the image  or are not in the ROI are looked at
                if x > baseROISize[0] or y > baseROISize[1] or z > baseROISize[2] or baseROI.GetPixel(x, y,
                                                                                                      z) != roiLabel:
                    # Once a pixel is confirmed to be outside the ROI, we get it's corresponding value from the baseImage
                    current_value = baseImage.GetPixel(x, y, z)
                    # Update the max and min values
                    maxVoxelVal = max(maxVoxelVal, current_value)
                    minVoxelVal = min(minVoxelVal, current_value)

    # Create a new base image so we are not directly editing the input image
    new_base = baseImage.__copy__()
    # Delete the input image to save space
    del (baseImage)

    # Now iterate over the pixels outside the ROI in the image and randomly generate a new value for them
    for x in range(baseImageSize[0]):
        for y in range(baseImageSize[1]):
            for z in range(baseImageSize[2]):
                if x > baseROISize[0] or y > baseROISize[1] or z > baseROISize[2] or baseROI.GetPixel(x, y,
                                                                                                      z) != roiLabel:
                    # Randomly assigning the current value to the range [maxVoxelVal, maxVoxelVal]
                    mapped_value = random.randint(minVoxelVal, maxVoxelVal)

                    # Set the new pixel value
                    new_base.SetPixel(x, y, z, mapped_value)

    return new_base


def shuffleNonROI(baseImage: sitk.Image, baseROI: sitk.Image, roiLabel: int = 1):
    """Function to shuffle all pixel values that are not within the Region of Interest in a sitk Image

    Parameters
    ----------
    baseImage : sitk.Image
        Image to shuffle the pixels outside the Region of Interest in
    baseROI : sitk.Image
        Image detailing Region of Interest
    roiLabel : int
        The label representing the ROI in baseROI

    Returns
    -------
    sitk.Image
        Image with all pixel values outside the Region of Interest randomly shuffled with same dimensions as input image
    """

    # A collection of corresponding value in the BaseImage of all the pixels not in the ROI
    count = []
    # Iterate through baseImage to store the corresponding value in the baseImage of all the pixels outside the ROI
    baseImageSize = baseImage.GetSize()
    baseROISize = baseROI.GetSize()
    for x in range(baseImageSize[0]):
        for y in range(baseImageSize[1]):
            for z in range(baseImageSize[2]):
                if x > baseROISize[0] or y > baseROISize[1] or z > baseROISize[2] or baseROI.GetPixel(x, y,
                                                                                                      z) != roiLabel:
                    # Here the key is the pixel coordinate and the value is the value of the pixel in the base image
                    count.append(baseImage.GetPixel(x, y, z))

    # Create a new base image so we are not directly editing the input image
    new_base = baseImage.__copy__()
    # Delete the input image to save space
    del (baseImage)

    # Randomly shuffling the pixel values
    random.shuffle(count)

    for x in range(baseImageSize[0]):
        for y in range(baseImageSize[1]):
            for z in range(baseImageSize[2]):
                if x > baseROISize[0] or y > baseROISize[1] or z > baseROISize[2] or baseROI.GetPixel(x, y,
                                                                                                      z) != roiLabel:
                    # Set the value of a pixel outside the ROI to be a shuffled value
                    new_base.SetPixel(x, y, z, count.pop())

    return new_base

def randomizeImageFromDistribtutionSampling(imageToRandomize: sitk.Image):
    """Function to randomly sample all the pixel values in a sitk Image, from the distribution of existing values

    Parameters
    ----------
    imageToRandomize : sitk.Image
        Image to randomly sample the pixels in

    Returns
    -------
    sitk.Image
        Image with all pixel values randomly sampled from the initial dstribution of the image, with same dimensions as input image
    """
    # Convert the image to an array
    arrImage = sitk.GetArrayFromImage(imageToRandomize)

    # Get array dimensions to reshape back to
    imgDimensions = arrImage.shape

    # Flatten the 3D array to 1D so values can be shuffled
    flatArrImage = arrImage.flatten()

    # Randomly sample values for new array from original image distribution
    sampled_array = np.random.choice(flatArrImage, size=len(flatArrImage), replace=True)

    # Reshape the array back into the original image dimensions
    randomlySampled3DArrImage = np.reshape(sampled_array, imgDimensions)

    # Convert back to sitk Image
    randomlySampledImage = sitk.GetImageFromArray(randomlySampled3DArrImage)

    # Set the origin/direction/spacing from original image to sampled image
    alignedRandomlySampledImage = alignImages(imageToRandomize, randomlySampledImage)

    return alignedRandomlySampledImage

def makeRandomFromRoiDistribution(baseImage: sitk.Image, baseROI: sitk.Image, roiLabel: int = 1):
    """Function to randomly sample pixel values within the Region of Interest uniformly from the distribution of pixel values in the ROI region sitk Image

    Parameters
    ----------
    baseImage : sitk.Image
        Image to randomly generate pixel values in Region of Interest
    baseROI : sitk.Image
        Image detailing Region of Interest
    roiLabel : int
        The label representing the ROI in baseROI

    Returns
    -------
    sitk.Image
        Image with all pixel values within the Region of Interest randomly sampled with same dimensions as input image
    """
    # Initialize array of ROI pixel distribution
    distributionROI = []

    # Iterate through baseROI to find the highest and lowest values in baseImage's ROI
    baseROISize = baseROI.GetSize()
    for x in range(baseROISize[0]):
        for y in range(baseROISize[1]):
            for z in range(baseROISize[2]):
                # Ensure only pixels that match the roiLabel of hte image are looked at
                if baseROI.GetPixel(x, y, z) == roiLabel:
                    # Once a pixel is confirmed to be in the ROI get it's corresponding value from the baseImage
                    distributionROI.append(baseImage.GetPixel(x, y, z))

    # Create a new base image so we are not directly editing the input image
    new_base = baseImage.__copy__()
    # Delete the input image to save space
    del (baseImage)

    # Now iterate over the pixels of the ROI in the image and randomly generate a new value for them
    for x in range(baseROISize[0]):
        for y in range(baseROISize[1]):
            for z in range(baseROISize[2]):
                if baseROI.GetPixel(x, y, z) == roiLabel:
                    # Assigning the current value to the randomly sampled value from within the ROI
                    mapped_value = random.choice(distributionROI)

                    # Set the new pixel value
                    new_base.SetPixel(x, y, z, mapped_value)

    return new_base

def makeRandomNonRoiFromDistribution(baseImage: sitk.Image, baseROI: sitk.Image, roiLabel: int = 1):
    """Function to random sample pixel values outside the Region of Interest uniformly from the distribution of pixel values outside the ROI in a sitk Image

    Parameters
    ----------
    baseImage : sitk.Image
        Image to randomly generate pixel values in outside the Region of Interest
    baseROI : sitk.Image
        Image detailing the Region of Interest
    roiLabel : int
        The label representing the ROI in baseROI

    Returns
    -------
    sitk.Image
        Image with all pixel values outside the Region of Interest randomly sample form outside the ROI with same dimensions as input image
    """

    # Initialize array of non-ROI pixel distribution
    distributionROI = []

    # Iterate through baseImage to find the highest and lowest values not in baseImage's ROI
    baseImageSize = baseImage.GetSize()
    baseROISize = baseROI.GetSize()
    for x in range(baseImageSize[0]):
        for y in range(baseImageSize[1]):
            for z in range(baseImageSize[2]):
                # Ensure only pixels that do not match the roiLabel of the image  or are not in the ROI are looked at
                if x > baseROISize[0] or y > baseROISize[1] or z > baseROISize[2] or baseROI.GetPixel(x, y,
                                                                                                      z) != roiLabel:
                    # Once a pixel is confirmed to be outside the ROI, we get it's corresponding value from the baseImage
                    distributionROI.append(baseImage.GetPixel(x, y, z))

    # Create a new base image so we are not directly editing the input image
    new_base = baseImage.__copy__()
    # Delete the input image to save space
    del (baseImage)

    # Now iterate over the pixels outside the ROI in the image and randomly generate a new value for them
    for x in range(baseImageSize[0]):
        for y in range(baseImageSize[1]):
            for z in range(baseImageSize[2]):
                if x > baseROISize[0] or y > baseROISize[1] or z > baseROISize[2] or baseROI.GetPixel(x, y,
                                                                                                      z) != roiLabel:
                    # Assigning the current value to the randomly sampled value from within the ROI
                    mapped_value = random.choice(distributionROI)

                    # Set the new pixel value
                    new_base.SetPixel(x, y, z, mapped_value)

    return new_base


def applyNegativeControl(nc_type: str, baseImage: sitk.Image, baseROI: sitk.Image = None, roiLabel: int = 1):
    """Function to generate random pixel values within the Region of Interest based on the range of values in a sitk Image

    Parameters
    ----------
    nc_type : str
        The type of negative control to be applied
    baseImage : sitk.Image
        The image to be modified
    baseROI : sitk.Image
        Image detailing Region of Interest
    roiLabel : int
        The label representing the ROI in baseROI

    Returns
    -------
    sitk.Image
        The output image with the negative control applied
     """

    if nc_type == "randomized_full":
        # Make negative control version of ctImage (randomized pixel size)
        return makeRandomImage(baseImage)
    elif nc_type == "randomized_roi":
        # Make negative control version of ctImage (randomized pixel size inside the ROI)
        return makeRandomRoi(baseImage, baseROI, roiLabel)
    elif nc_type == "shuffled_roi":
        # Make negative control version of ctImage (random shuffled pixels inside the ROI, same size)
        return shuffleROI(baseImage, baseROI, roiLabel)
    elif nc_type == "randomized_non_roi":
        # Make negative control version of ctImage (randomized pixel size outside the ROI)
        return makeRandomNonRoi(baseImage, baseROI, roiLabel)
    elif nc_type == "shuffled_non_roi":
        # Make negative control version of ctImage (shuffled pixels outside the ROI, same size)
        return shuffleNonROI(baseImage, baseROI, roiLabel)
    elif nc_type == "randomized_sampled_full":
        # Make negative control version of ctImage (random sampled pixels from original distribution, same size)
        return randomizeImageFromDistribtutionSampling(baseImage)
    elif nc_type == "randomized_sampled_roi":
        # Make negative control version of ctImage (random sampled pixels from original distribution inside ROI, same size)
        return makeRandomFromRoiDistribution(baseImage, baseROI, roiLabel)
    elif nc_type == "randomized_sampled_non_roi":
        # Make negative control version of ctImage (random sampled pixels from original distribution outside ROI, same size)
        return makeRandomNonRoiFromDistribution(baseImage, baseROI, roiLabel)
    elif nc_type == "shuffled_full":
        # Make negative control version of ctImage (random shuffled pixels, same size)
        return shuffleImage(baseImage)
